var querystring = require('querystring');
var https = require('https');
var fs = require('fs');

var loadToken = function(host, callback) {
  // Try reading C:\conduit_token.txt, or use ~/.arcrc.
  console.log('phabricator: Checking if Conduit token is present at C:\\conduit_token.txt...');
  fs.access('C:\\conduit_token.txt', fs.constants.R_OK, (err) => {
    if (err) {
      console.log('phabricator: Checking if Conduit token is present at ~/.arcrc...');
      fs.access(process.env.HOME + '/.arcrc', fs.constants.R_OK, (err) => {
        if (err) {
          callback("Unable to load Conduit token from either C:\conduit_token.txt or ~/.arcrc!", null);
        } else {
          console.log('phabricator: Reading Conduit token from ~/.arcrc...');
          fs.readFile(process.env.HOME + '/.arcrc', 'utf8', (err, data) => {
            if (err) {
              console.log('phabricator: Failed to read token from ~/.arcrc: ' + err);
              callback(err, null);
            } else {
              var token = JSON.parse(data)['hosts']['https://'+host+'/api/']['token'];
              console.log('phabricator: Read Conduit token from ~/.arcrc.');
              callback(null, token);
            }
          });
        }
      });
    } else {
      console.log('phabricator: Reading Conduit token from C:\\conduit_token.txt...');
      fs.readFile('C:\\conduit_token.txt', 'utf8', (err, data) => {
        if (err) {
          console.log('phabricator: Failed to read token from C:\\conduit_token.txt: ' + err);
          callback(err, null);
        } else {
          var token = data.trim();
          console.log('phabricator: Read Conduit token from C:\\conduit_token.txt.');
          callback(null, token);
        }
      });
    }
  });
}

var invokeRawConduitMethod = function(host, method, params, callback) {
  console.log('phabricator: Loading Conduit token in preparation for request...');
  loadToken(host, (err, token) => {
    if (err) {
      console.log('phabricator: Failed to obtain Conduit token: ' + err);
      callback(err, null);
      return;
    }

    // Create the POST data.
    var paramsClone = JSON.parse(JSON.stringify(params));
    paramsClone['__conduit__'] = { token: token };
    var postData = querystring.stringify({ params: JSON.stringify(paramsClone) });

    // Configure the HTTPS request.
    var options = {
      host: host,
      path: '/api/' + method,
      port: 443,
      method: 'POST',
      headers: {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Content-Length': Buffer.byteLength(postData), 
        'Accept': '*/*'
      },
    };

    // Define the attempt function so we can retry up to 5 times.
    console.log('phabricator: Making HTTPS request to https://' + options.host + options.path + '...');
    var performAttempt = function(attempt, callback) {
      console.log('phabricator: Attempting request to ' + method + ' (' + attempt + '/5)...');
      var req = https.request(options, (res) => {
        console.log('phabricator: Retrieving response data...');
        res.setEncoding('utf8');
        var data = "";
        res.on('data', (chunk) => {
          data += chunk;
        });
        res.on('end', () => {
          try {
            var results = JSON.parse(data);
            console.log('phabricator: Obtained response from Conduit.');
            callback(null, results);
            return;
          } catch (err) {
            console.log('phabricator: Failed to retrieve response data: ' + err);
            if (attempt == 5) {
              callback(err, null);
              return;
            } else {
              performAttempt(attempt + 1, callback);
            }
          }
        });
      });
      req.on('error', (err) => {
        console.log('phabricator: Failed to initiate request: ' + err);
        if (attempt == 5) {
          callback(err, null);
          return;
        } else {
          performAttempt(attempt + 1, callback);
        }
      });
      req.write(postData);
      req.end();
    };

    // Make the first attempt.
    performAttempt(1, callback);
  });
};

var getCredentials = function(host, ids, callback) {
  invokeRawConduitMethod(host, 'passphrase.query', {'ids': ids, 'needSecrets': true}, (err, response) => {
    if (err) {
      callback(err, null);
    } else {
      if (response.error_code != null) {
        callback(response.error_code + ": " + response.error_info, null);
        return;
      }

      var results = {};
      for (var key in response.result.data) {
        var phid = key;
        var data = response.result.data[key];

        results[data.id] = {
          username: data.username,
          password: data.material.password
        };
      }

      callback(null, results);
    }
  });
}

var getCredentialsFromIdList = function(host, id_list, callback) {
  var ids_to_fetch = [];
  for (var key in id_list) {
    if (id_list.hasOwnProperty(key)) {
      for (var key2 in id_list[key]) {
        if (id_list[key].hasOwnProperty(key2)) {
          var id = id_list[key][key2].id;
          if (ids_to_fetch.indexOf(id) == -1) {
            ids_to_fetch.push(id);
          }
        }
      }
    }
  }

  getCredentials(host, ids_to_fetch, (err, result) => {
    if (err) {
      callback(err, null);
    } else {
      var clone = JSON.parse(JSON.stringify(id_list));
      for (var key in clone) {
        if (clone.hasOwnProperty(key)) {
          for (var key2 in clone[key]) {
            if (clone[key].hasOwnProperty(key2)) {
              var id = clone[key][key2].id;
              if (clone[key][key2].type == 'password') {
                clone[key][key2] = result[id].password;
              } else {
                clone[key][key2] = result[id].username;
              }
            }
          }
        }
      }
      callback(null, clone);
    }
  });
}

module.exports = {
  loadToken: loadToken,
  invokeRawConduitMethod: invokeRawConduitMethod,
  getCredentials: getCredentials,
  getCredentialsFromIdList: getCredentialsFromIdList
}